use base64::{EncodeSliceError, prelude::*};
use core::str::from_utf8;
use core::{fmt::Write, str::FromStr};
use embassy_net::tcp::TcpSocket;
use embassy_time::{Duration, Timer};
use embedded_io_async::Write as _;
use heapless::{String, Vec};
use log::{info, warn};
use sha1::{Digest, Sha1};

use crate::{apps, socket::ws::Ws};

pub mod ws;

#[cfg(feature = "ttt")]
#[embassy_executor::task(pool_size = 2)]
pub async fn ttt_listen_task(stack: embassy_net::Stack<'static>, team: apps::ttt::Team, port: u16) {
    listen_task(stack, apps::ttt::TttApp::new(team), port).await
}

#[embassy_executor::task(pool_size = 2)]
pub async fn index_listen_task(stack: embassy_net::Stack<'static>, port: u16) {
    listen_task(stack, apps::index::IndexApp, port).await
}

#[cfg(feature = "chat")]
#[embassy_executor::task(pool_size = 4)]
pub async fn chat_listen_task(stack: embassy_net::Stack<'static>, port: u16) {
    listen_task(stack, apps::chat::ChatApp::new(), port).await
}

pub async fn listen_task(stack: embassy_net::Stack<'static>, mut app: impl apps::App, port: u16) {
    // loop {
    //     info!("team:{:?}", team);
    //     Timer::after_millis(0).await;
    // }
    let mut rx_buffer = [0; 1024];
    let mut tx_buffer = [0; 2048];
    let mut buf = [0; 1024];
    let mut head_buf = Vec::<u8, 256>::new();
    loop {
        Timer::after_secs(0).await;
        let mut socket = TcpSocket::new(stack, &mut rx_buffer, &mut tx_buffer);
        socket.set_timeout(Some(Duration::from_secs(30)));

        info!("Socket {}: Listening on TCP:{}...", app.socket_name(), port);
        if let Err(e) = socket.accept(port).await {
            warn!("accept error: {:?}", e);
            continue;
        }

        info!(
            "Socket {}: Received connection from {:?}",
            app.socket_name(),
            socket.remote_endpoint()
        );

        loop {
            Timer::after_secs(0).await;
            let n = match socket.read(&mut buf).await {
                Ok(0) => {
                    warn!("read EOF");
                    break;
                }
                Ok(n) => n,
                Err(e) => {
                    warn!("Socket {}: read error: {:?}", app.socket_name(), e);
                    break;
                }
            };
            head_buf.clear();
            let (headers, content) = match from_utf8(&buf[..n]) {
                Ok(b) => match b.split_once("\r\n\r\n") {
                    Some(t) => t,
                    None => (b, ""),
                },
                Err(_) => {
                    warn!("Non utf8 http request");
                    break;
                }
            };

            info!("\n{:?}\n", headers);

            let mut hl = headers.lines();
            let (request_type, path) = match hl.next() {
                None => {
                    warn!("Empty request");
                    break;
                }
                Some(l1) => {
                    let mut l1 = l1.split(' ');
                    (
                        match l1.next() {
                            Some("GET") => HttpRequestType::Get,
                            Some("POST") => HttpRequestType::Post,
                            Some(t) => {
                                warn!("Unknown request type : {}", t);
                                break;
                            }
                            None => {
                                warn!("No request type");
                                break;
                            }
                        },
                        match l1.next() {
                            Some(path) => path,
                            None => {
                                warn!("No path");
                                break;
                            }
                        },
                    )
                }
            };
            let mut host = None;
            let mut ws_handshake = false;
            let mut ws_key = None;
            for h in hl {
                let Some((name, val)) = h.split_once(':') else {
                    continue;
                };
                let name = name.trim();
                let val = val.trim();
                match (name, val) {
                    ("Host", _) => host = Some(val),
                    ("Upgrade", "websocket") => ws_handshake = true,
                    ("Sec-WebSocket-Key", _) => ws_key = Some(val),
                    _ => {}
                }
            }
            let Some(host) = host else {
                warn!("No host");
                break;
            };

            info!(
                "Socket {}: {:?}{} request for {}{}",
                app.socket_name(),
                request_type,
                if ws_handshake { " websocket" } else { "" },
                host,
                path,
            );
            Timer::after_secs(0).await;

            head_buf.clear();
            let res_content: Result<&str, core::fmt::Error> = try {
                if ws_handshake {
                    if !app.accept_ws(path) {
                        write!(
                            &mut head_buf,
                            "{}\r\n\r\n",
                            Into::<&str>::into(HttpResCode::NotFound)
                        )?;
                        ""
                    } else {
                        if path.len() > 16 {
                            warn!("Ws socket cannot have path longer than 16 chars!");
                            break;
                        }
                        let Some(key) = ws_key else {
                            warn!("No ws key!");
                            break;
                        };
                        let accept = match compute_ws_accept(key).await {
                            Ok(a) => a,
                            Err(e) => {
                                warn!("compute ws accept error : {:?}", e);
                                break;
                            }
                        };
                        write!(
                            &mut head_buf,
                            "{}\r\n\
                            Upgrade: websocket\r\n\
                            Connection: Upgrade\r\n\
                            Sec-WebSocket-Accept: {}\r\n\r\n",
                            // Sec-WebSocket-Protocol: chat\r\n
                            Into::<&str>::into(HttpResCode::SwitchingProtocols),
                            accept
                        )?;
                        ""
                    }
                } else {
                    let (code, res_type, res_content): (HttpResCode, &str, &str) = match path {
                        "/htmx.js" => (
                            HttpResCode::Ok,
                            "javascript",
                            #[cfg(debug_assertions)]
                            include_str!("../static/htmx.js"),
                            #[cfg(not(debug_assertions))]
                            include_bytes!("../static/htmx.min.js"),
                        ),
                        _ => app.handle_request(path, request_type, content).await,
                    };

                    write!(&mut head_buf, "{}", Into::<&str>::into(code))?;
                    if res_type.len() > 0 {
                        write!(
                            &mut head_buf,
                            "\r\n\
                                Content-Type: text/{}\r\n\
                                Content-Length: {}\r\n",
                            res_type,
                            res_content.len()
                        )?;
                    }
                    write!(&mut head_buf, "\r\n\r\n")?;
                    res_content
                }
            };

            let res_content = match res_content {
                Ok(rc) => rc,
                Err(e) => {
                    warn!("res buffer write error: {:?}", e);
                    break;
                }
            };

            info!("\n{}\n", from_utf8(&head_buf).unwrap());

            match socket.write_all(&head_buf).await {
                Ok(()) => {}
                Err(e) => {
                    warn!("write error: {:?}", e);
                    break;
                }
            };
            match socket.write_all(res_content.as_bytes()).await {
                Ok(()) => {}
                Err(e) => {
                    warn!("write error: {:?}", e);
                    break;
                }
            };

            if ws_handshake {
                let path: String<16> = String::from_str(path).unwrap();
                app.handle_ws(
                    &path,
                    Ws::new(&mut socket, &mut buf, &mut head_buf, app.socket_name()),
                )
                .await;
                break;
            }
        }
    }
}

#[derive(Clone, Copy, Debug)]
pub enum HttpRequestType {
    Get,
    Post,
}
impl Into<&str> for HttpRequestType {
    fn into(self) -> &'static str {
        match self {
            Self::Get => "GET",
            Self::Post => "POST",
        }
    }
}

#[derive(Debug, Clone, Copy)]
pub enum HttpResCode {
    SwitchingProtocols,
    Ok,
    NoContent,
    BadRequest,
    NotFound,
    Forbidden,
}
impl Into<&str> for HttpResCode {
    fn into(self) -> &'static str {
        match self {
            HttpResCode::SwitchingProtocols => "HTTP/1.1 101 Switching Protocols",
            HttpResCode::Ok => "HTTP/1.1 200 OK",
            HttpResCode::NoContent => "HTTP/1.1 204 NO CONTENT",
            HttpResCode::BadRequest => "HTTP/1.1 400 BAD REQUEST",
            HttpResCode::NotFound => "HTTP/1.1 404 NOT FOUND",
            HttpResCode::Forbidden => "HTTP/1.1 403 FORBIDDEN",
        }
    }
}

async fn compute_ws_accept(key: &str) -> Result<String<28>, EncodeSliceError> {
    let mut res = Vec::<u8, 28>::new();
    res.extend_from_slice(&[0; 28]).unwrap();
    let mut hasher = Sha1::new();
    hasher.update(key.as_bytes());
    hasher.update(b"258EAFA5-E914-47DA-95CA-C5AB0DC85B11");
    let hash = hasher.finalize();
    BASE64_STANDARD.encode_slice(hash, &mut res)?;
    Ok(String::from_utf8(res).unwrap())
}
