use core::str::from_utf8;

use embassy_net::tcp::{TcpReader, TcpSocket, TcpWriter};
use embassy_time::Instant;
use embedded_io_async::{ErrorType, ReadReady, Write};
use heapless::Vec;
use log::{info, warn};

#[derive(Clone, Copy)]
pub enum WsMsg<'a> {
    Ping(&'a [u8]),
    Pong(&'a [u8]),
    Text(&'a str),
    Bytes(&'a [u8]),
    Unknown(u8, &'a [u8]),
}
impl WsMsg<'_> {
    pub const fn len(&self) -> usize {
        self.as_bytes().len()
    }
    pub const fn as_bytes(&self) -> &[u8] {
        match self {
            WsMsg::Text(t) => t.as_bytes(),
            WsMsg::Bytes(b) | WsMsg::Pong(b) | WsMsg::Ping(b) | WsMsg::Unknown(_, b) => b,
        }
    }
    pub const fn code(&self) -> u8 {
        match self {
            WsMsg::Text(_) => 1,
            WsMsg::Bytes(_) => 2,
            WsMsg::Ping(_) => 9,
            WsMsg::Pong(_) => 10,
            WsMsg::Unknown(c, _) => *c,
        }
    }
}

struct WsRx<'a, const BUF_SIZE: usize> {
    socket: TcpReader<'a>,
    buf: &'a mut [u8; BUF_SIZE],
    last_msg: Instant,
}
struct WsTx<'a, const HEAD_BUF_SIZE: usize> {
    socket: TcpWriter<'a>,
    head_buf: &'a mut Vec<u8, HEAD_BUF_SIZE>,
}
impl<'a, const HEAD_BUF_SIZE: usize> WsTx<'a, HEAD_BUF_SIZE> {
    pub async fn send<'m>(&mut self, msg: WsMsg<'m>) -> Result<(), ()> {
        self.head_buf.clear();
        self.head_buf.push(0b1000_0000 | msg.code()).unwrap();
        if msg.len() < 126 {
            self.head_buf.push(msg.len() as u8).unwrap();
        } else {
            self.head_buf.push(0b0111_1110).unwrap();
            self.head_buf
                .extend_from_slice(&(msg.len() as u16).to_le_bytes())
                .unwrap();
            self.head_buf.extend_from_slice(msg.as_bytes()).unwrap();
        }
        let w: Result<(), <TcpSocket<'_> as ErrorType>::Error> = try {
            self.socket.write_all(&self.head_buf).await?;
            self.socket.write_all(msg.as_bytes()).await?;
        };
        w.map_err(|e| {
            warn!("write error: {:?}", e);
            ()
        })
    }
}

pub struct Ws<'a, const BUF_SIZE: usize = 1024, const RES_HEAD_BUF_SIZE: usize = 256> {
    rx: WsRx<'a, BUF_SIZE>,
    tx: WsTx<'a, RES_HEAD_BUF_SIZE>,
    name: &'a str,
}
impl<'a, const BUF_SIZE: usize, const HEAD_BUF_SIZE: usize> Ws<'a, BUF_SIZE, HEAD_BUF_SIZE> {
    pub fn new(
        socket: &'a mut TcpSocket,
        buf: &'a mut [u8; BUF_SIZE],
        head_buf: &'a mut Vec<u8, HEAD_BUF_SIZE>,
        name: &'a str,
    ) -> Self {
        let (rx, tx) = socket.split();
        Self {
            rx: WsRx {
                socket: rx,
                buf,
                last_msg: Instant::MIN,
            },
            tx: WsTx {
                socket: tx,
                head_buf,
            },
            name,
        }
    }
    // Do this often to respond to pings
    async fn rcv(&mut self) -> Result<Option<WsMsg>, ()> {
        if !self.rx.socket.read_ready().unwrap() {
            return Ok(None);
        }
        let n = match self.rx.socket.read(self.rx.buf).await {
            Ok(0) => {
                warn!("read EOF");
                return Err(());
            }
            Ok(n) => n,
            Err(e) => {
                warn!("Socket {}: read error: {:?}", self.name, e);
                return Err(());
            }
        };
        if self.rx.buf[0] & 0b1000_0000 == 0 {
            warn!("Fragmented ws messages are not supported!");
            return Err(());
        }
        if self.rx.buf[0] & 0b0111_0000 != 0 {
            warn!(
                "Reserved ws bits are set : {}",
                (self.rx.buf[0] >> 4) & 0b0111
            );
            return Err(());
        }
        let (length, n_after_length) = match self.rx.buf[1] & 0b0111_1111 {
            126 => (
                u64::from_le_bytes([0, 0, 0, 0, 0, 0, self.rx.buf[2], self.rx.buf[3]]),
                4,
            ),
            127 => (
                u64::from_le_bytes(self.rx.buf[2..10].try_into().unwrap()),
                10,
            ),
            l => (l as u64, 2),
        };
        if length > 512 {
            warn!("ws payload bigger than 512!");
            return Err(());
        }

        let content = if self.rx.buf[1] & 0b1000_0000 != 0 {
            // masked message
            if n_after_length + 4 + length as usize > n {
                warn!("ws payload smaller than length");
                return Err(());
            }
            let mask_key: [u8; 4] = self.rx.buf[n_after_length..n_after_length + 4]
                .try_into()
                .unwrap();
            for (i, x) in self.rx.buf[n_after_length + 4..n_after_length + 4 + length as usize]
                .iter_mut()
                .enumerate()
            {
                *x ^= mask_key[i & 0xff];
            }
            &self.rx.buf[n_after_length + 4..n_after_length + 4 + length as usize]
        } else {
            if n_after_length + length as usize > n {
                warn!("ws payload smaller than length");
                return Err(());
            }
            &self.rx.buf[n_after_length..n_after_length + length as usize]
        };
        self.rx.last_msg = Instant::now();
        match self.rx.buf[0] & 0b0000_1111 {
            // Text  message
            1 => {
                let content = from_utf8(&content).map_err(|_| ())?;
                info!("Received text : {:?}", content);
                Ok(Some(WsMsg::Text(content)))
            }
            // Ping
            9 => {
                self.tx.send(WsMsg::Pong(&content)).await?;
                Ok(Some(WsMsg::Ping(&content)))
            }
            // Pong
            10 => Ok(Some(WsMsg::Pong(&content))),
            c => {
                info!("Unknown ws op code (ignoring) : {}", c);
                Ok(Some(WsMsg::Unknown(c, &content)))
            }
        }
    }
    pub async fn send(&mut self, msg: WsMsg<'a>) -> Result<(), ()> {
        self.tx.send(msg).await
    }
}
